-module(kv_ring).

-export([
    new/2,
    key_to_index/1,
    ordered_from/2
]).

-define(RING_MAX, trunc(math:pow(2, 160))).

-export_type([ring/0]).

-type index() :: non_neg_integer().

-opaque ring() :: {Partitions :: non_neg_integer(), [{Index :: index(), Owner :: node()}]}.

%%% public api

-spec new(Partitions :: non_neg_integer(), Node :: node()) -> ring().
new(Partitions, Node) ->
    {Partitions, [{Index, Node} || Index <- lists:seq(0, ?RING_MAX - 1, ring_step(Partitions))]}.

-spec key_to_index(Key :: binary()) -> index().
key_to_index(Key) ->
    <<Index:160/integer>> = crypto:hash(sha, Key),
    Index.

-spec ordered_from(Ring :: ring(), Index :: index()) -> [{Index :: index(), Owner :: node()}].
ordered_from({Partitions, OwnerList}, Index) ->
    Step = ring_step(Partitions),
    {A, B} = lists:split(Index div Step, OwnerList),
    B ++ A.

%%% internal functions

ring_step(RingSize) ->
    ?RING_MAX div RingSize.
