-module(kv_vnode_manager).

-behaviour(gen_server).

-export([
    start_link/0,
    get_local/1,
    put_local/2
]).

%%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2]).

-type state() :: {}.

%%% public api

start_link() ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).

-spec get_local(Key :: binary()) -> {ok, Value :: term()} | none.
get_local(Key) ->
    Index = kv_ring:key_to_index(Key),
    VNode = get_vnode(Index),
    kv_vnode:get(VNode, Key).

-spec put_local(Key :: binary(), Value :: term()) -> ok.
put_local(Key, Value) ->
    Index = kv_ring:key_to_index(Key),
    VNode = get_vnode(Index),
    kv_vnode:put(VNode, Key, Value).

%%% gen_server callbacks

-spec init([]) -> {ok, state()}.
init([]) ->
    ?MODULE = ets:new(?MODULE, [set, protected, named_table, {read_concurrency, true}]),
    {ok, {}}.

handle_call({get_vnode, Index}, _From, State) ->
    VNode =
        case ets:lookup(?MODULE, Index) of
            [{_, Pid}] ->
                Pid;
            [] ->
                {ok, Pid} = kv_vnode_sup:start_vnode(Index),
                ets:insert(?MODULE, {Index, Pid}),
                Pid
        end,
    {reply, VNode, State}.

handle_cast(_, _) ->
    error(not_implemented).

%%% internal functions

-spec get_vnode(Index :: kv_ring:index()) -> pid().
get_vnode(Index) ->
    case ets:lookup(?MODULE, Index) of
        [{_, Pid}] ->
            Pid;
        [] ->
            gen_server:call(?MODULE, {get_vnode, Index})
    end.
